#include <biosnake/alignment/compressedA3M.h>
#include <biosnake/commons/dBConcat.h>
#include <biosnake/commons/dBReader.h>
#include <biosnake/commons/dBWriter.h>
#include <biosnake/output.h>
#include <biosnake/commons/headerSummarizer.h>
#include <biosnake/alignment/msaFilter.h>
#include <biosnake/alignment/pSSMCalculator.h>
#include <biosnake/commons/parameters.h>
#include <biosnake/commons/util.h>
#include <biosnake/output.h>

#ifdef OPENMP
#include <omp.h>
#endif

int result2msa(biosnake_output *out, Parameters &par) {
  // BiosnakeMPI::init(argc, argv);

  //    Parameters &par = Parameters::getInstance();
  //    // do not filter by default
  //    par.filterMsa = 0;
  //    par.pca = 0.0;
  //    par.parseParameters(argc, argv, command, true, 0, 0);

  const bool isCA3M =
      par.msaFormatMode == Parameters::FORMAT_MSA_CA3M ||
      par.msaFormatMode == Parameters::FORMAT_MSA_CA3M_CONSENSUS;
  const bool shouldWriteNullByte =
      par.msaFormatMode != Parameters::FORMAT_MSA_STOCKHOLM_FLAT;

  DBReader<unsigned int> qDbr(
      out, par.db1.c_str(), par.db1Index.c_str(), par.threads,
      DBReader<unsigned int>::USE_INDEX | DBReader<unsigned int>::USE_DATA);
  qDbr.open(DBReader<unsigned int>::NOSORT);
  DBReader<unsigned int> queryHeaderReader(
      out, par.hdr1.c_str(), par.hdr1Index.c_str(), par.threads,
      DBReader<unsigned int>::USE_INDEX | DBReader<unsigned int>::USE_DATA);
  queryHeaderReader.open(DBReader<unsigned int>::NOSORT);
  if (par.preloadMode != Parameters::PRELOAD_MODE_MMAP) {
    qDbr.readMmapedDataInMemory();
    queryHeaderReader.readMmapedDataInMemory();
  }

  DBReader<unsigned int> *tDbr = &qDbr;
  DBReader<unsigned int> *targetHeaderReader = &queryHeaderReader;
  unsigned int maxSequenceLength = qDbr.getMaxSeqLen();

  const bool sameDatabase = (par.db1.compare(par.db2) == 0) ? true : false;
  if (!sameDatabase) {
    tDbr = new DBReader<unsigned int>(
        out, par.db2.c_str(), par.db2Index.c_str(), par.threads,
        DBReader<unsigned int>::USE_INDEX | DBReader<unsigned int>::USE_DATA);
    tDbr->open(DBReader<unsigned int>::NOSORT);
    maxSequenceLength = std::max(qDbr.getMaxSeqLen(), tDbr->getMaxSeqLen());
    if (par.preloadMode != Parameters::PRELOAD_MODE_MMAP) {
      tDbr->readMmapedDataInMemory();
    }

    if (isCA3M == false) {
      targetHeaderReader = new DBReader<unsigned int>(
          out, par.hdr2.c_str(), par.hdr2Index.c_str(), par.threads,
          DBReader<unsigned int>::USE_INDEX | DBReader<unsigned int>::USE_DATA);
      targetHeaderReader->open(DBReader<unsigned int>::NOSORT);

      if (par.preloadMode != Parameters::PRELOAD_MODE_MMAP) {
        targetHeaderReader->readMmapedDataInMemory();
      }
    }
  }

  DBConcat *seqConcat = NULL;
  DBReader<unsigned int> *refReader = NULL;
  std::string outDb = par.db4;
  std::string outIndex = par.db4Index;
  if (isCA3M == true) {
    std::string refData = outDb + "_sequence.ffdata";
    std::string refIndex = outDb + "_sequence.ffindex";
    // Use only 1 thread for concat to ensure the same order
    seqConcat = new DBConcat(out, par.db1, par.db1Index, par.db2, par.db2Index,
                             refData, refIndex, 1, BiosnakeMPI::isMaster());
    DBConcat hdrConcat(out, par.hdr1, par.hdr1Index, par.hdr2, par.hdr2Index,
                       outDb + "_header.ffdata", outDb + "_header.ffindex", 1,
                       BiosnakeMPI::isMaster(), false, false, false, 2);

#ifdef HAVE_MPI
    MPI_Barrier(MPI_COMM_WORLD);
#endif
    // When exporting in ca3m, we need to access with SORT_BY_LINE
    // mode in order to keep track of the original line number in the index
    // file.
    refReader = new DBReader<unsigned int>(out, refData.c_str(), refIndex.c_str(),
                                           par.threads,
                                           DBReader<unsigned int>::USE_INDEX);
    refReader->open(DBReader<unsigned int>::SORT_BY_LINE);

    outDb = par.db4 + "_ca3m.ffdata";
    outIndex = par.db4 + "_ca3m.ffindex";
  }

  DBReader<unsigned int> resultReader(
      out, par.db3.c_str(), par.db3Index.c_str(), par.threads,
      DBReader<unsigned int>::USE_DATA | DBReader<unsigned int>::USE_INDEX);
  resultReader.open(DBReader<unsigned int>::LINEAR_ACCCESS);
  size_t dbFrom = 0;
  size_t dbSize = 0;
#ifdef HAVE_MPI
  resultReader.decomposeDomainByAminoAcid(BiosnakeMPI::rank, BiosnakeMPI::numProc,
                                          &dbFrom, &dbSize);
  out->info("Compute split from {}", dbFrom << " to "
                     << dbFrom + dbSize);
  std::pair<std::string, std::string> tmpOutput =
      Util::createTmpFileNames(outDb, outIndex, BiosnakeMPI::rank);
#else
  dbSize = resultReader.getSize();
  std::pair<std::string, std::string> tmpOutput =
      std::make_pair(outDb, outIndex);
#endif

  int localThreads = par.threads;
  if (static_cast<int>(resultReader.getSize()) <= par.threads) {
    localThreads = static_cast<int>(resultReader.getSize());
  }

  size_t mode = par.compressed;
  int type = Parameters::DBTYPE_MSA_DB;
  if (isCA3M == true) {
    mode |= Parameters::WRITER_LEXICOGRAPHIC_MODE;
    type = Parameters::DBTYPE_CA3M_DB;
  } else if (par.msaFormatMode == Parameters::FORMAT_MSA_STOCKHOLM_FLAT) {
    type = Parameters::DBTYPE_OMIT_FILE;
  }
  DBWriter resultWriter(out, tmpOutput.first.c_str(), tmpOutput.second.c_str(),
                        localThreads, mode, type);
  resultWriter.open();

  // + 1 for query
  size_t maxSetSize = resultReader.maxCount('\n') + 1;

  // adjust score of each match state by -0.2 to trim alignment
  SubstitutionMatrix subMat(out, par.scoringMatrixFile.aminoacids, 2.0f, -0.2f);
  EvalueComputation evalueComputation(out, tDbr->getAminoAcidDBSize(), &subMat,
                                      par.gapOpen.aminoacids,
                                      par.gapExtend.aminoacids);
  if (qDbr.getDbtype() == -1 || tDbr->getDbtype() == -1) {
    out->failure("Please recreate your database or add a .dbtype file to your sequence/profile database");
    return EXIT_FAILURE;
  }
  if (Parameters::isEqualDbtype(qDbr.getDbtype(),
                                Parameters::DBTYPE_HMM_PROFILE) &&
      Parameters::isEqualDbtype(tDbr->getDbtype(),
                                Parameters::DBTYPE_HMM_PROFILE)) {
    out->failure("Only the query OR the target database can be a profile database");
    return EXIT_FAILURE;
  }
  out->info("Query database size: {} type: {}. Target database size: {} type: {}", qDbr.getSize(), qDbr.getDbTypeName(), tDbr->getSize(), tDbr->getDbTypeName());

  const bool isFiltering = par.filterMsa != 0;
  Log::Progress progress(dbSize - dbFrom);
#pragma omp parallel num_threads(localThreads)
  {
    unsigned int thread_idx = 0;
#ifdef OPENMP
    thread_idx = (unsigned int)omp_get_thread_num();
#endif

    Matcher matcher(out, qDbr.getDbtype(), maxSequenceLength, &subMat,
                    &evalueComputation, par.compBiasCorrection,
                    par.gapOpen.aminoacids, par.gapExtend.aminoacids);
    MultipleAlignment aligner(out, maxSequenceLength, &subMat);
    PSSMCalculator calculator(out, &subMat, maxSequenceLength, maxSetSize, par.pca,
                              par.pcb);
    MsaFilter filter(out, maxSequenceLength, maxSetSize, &subMat,
                     par.gapOpen.aminoacids, par.gapExtend.aminoacids);
    UniprotHeaderSummarizer summarizer;
    Sequence centerSequence(out, maxSequenceLength, qDbr.getDbtype(), &subMat, 0,
                            false, par.compBiasCorrection);
    Sequence edgeSequence(out, maxSequenceLength, tDbr->getDbtype(), &subMat, 0,
                          false, false);

    // which sequences where kept after filtering
    bool *kept = new bool[maxSetSize];
    for (size_t i = 0; i < maxSetSize; ++i) {
      kept[i] = 1;
    }

    char dbKey[255];
    const char *entry[255];
    std::string accession;

    std::vector<std::string> headers;
    headers.reserve(300);

    std::vector<Matcher::result_t> alnResults;
    alnResults.reserve(300);

    std::vector<std::vector<unsigned char>> seqSet;
    seqSet.reserve(300);

    std::vector<size_t> seqIds;
    seqIds.reserve(300);

    std::string result;
    result.reserve(300 * 1024);

#pragma omp for schedule(dynamic, 10)
    for (size_t id = dbFrom; id < (dbFrom + dbSize); id++) {
      progress.updateProgress();

      unsigned int queryKey = resultReader.getDbKey(id);
      size_t queryId = qDbr.getId(queryKey);
      if (queryId == UINT_MAX) {
        out->warn("Invalid query sequence {}", queryKey);
        continue;
      }
      centerSequence.mapSequence(queryId, queryKey,
                                 qDbr.getData(queryId, thread_idx),
                                 qDbr.getSeqLen(queryId));

      size_t centerHeaderId = queryHeaderReader.getId(queryKey);
      if (centerHeaderId == UINT_MAX) {
        out->warn("Invalid query header {}", queryKey);
        continue;
      }
      char *centerSequenceHeader =
          queryHeaderReader.getData(centerHeaderId, thread_idx);
      size_t centerHeaderLength =
          queryHeaderReader.getEntryLen(centerHeaderId) - 1;

      if (par.msaFormatMode == Parameters::FORMAT_MSA_STOCKHOLM_FLAT) {
        accession = Util::parseFastaHeader(centerSequenceHeader);
      }

      bool isQueryInit = false;
      char *data = resultReader.getData(id, thread_idx);
      while (*data != '\0') {
        Util::parseKey(data, dbKey);
        const unsigned int key = (unsigned int)strtoul(dbKey, NULL, 10);
        // in the same database case, we have the query repeated
        if (key == queryKey && sameDatabase == true) {
          data = Util::skipLine(data);
          continue;
        }

        const size_t edgeId = tDbr->getId(key);
        if (edgeId == UINT_MAX) {
          out->failure("Sequence {} does not exist in target sequence database", key);
        }
        edgeSequence.mapSequence(edgeId, key, tDbr->getData(edgeId, thread_idx),
                                 tDbr->getSeqLen(edgeId));
        seqSet.emplace_back(std::vector<unsigned char>(
            edgeSequence.numSequence,
            edgeSequence.numSequence + edgeSequence.L));
        seqIds.emplace_back(edgeId);

        const size_t columns = Util::getWordsOfLine(data, entry, 255);
        if (columns > Matcher::ALN_RES_WITHOUT_BT_COL_CNT) {
          alnResults.emplace_back(Matcher::parseAlignmentRecord(out, data));
        } else {
          // Recompute if not all the backtraces are present
          if (isQueryInit == false) {
            matcher.initQuery(&centerSequence);
            isQueryInit = true;
          }
          alnResults.emplace_back(
              matcher.getSWResult(&edgeSequence, INT_MAX, false, 0, 0.0,
                                  FLT_MAX, Matcher::SCORE_COV_SEQID, 0, false));
        }
        data = Util::skipLine(data);
      }

      MultipleAlignment::MSAResult res = aligner.computeMSA(
          &centerSequence, seqSet, alnResults, !par.allowDeletion);
      // MultipleAlignment::print(res, &subMat);

      if (par.msaFormatMode == Parameters::FORMAT_MSA_FASTADB ||
          par.msaFormatMode == Parameters::FORMAT_MSA_FASTADB_SUMMARY) {
        if (isFiltering) {
          filter.filter(res.setSize, res.centerLength,
                        static_cast<int>(par.covMSAThr * 100),
                        static_cast<int>(par.qid * 100), par.qsc,
                        static_cast<int>(par.filterMaxSeqId * 100), par.Ndiff,
                        (const char **)res.msaSequence, false);
          filter.getKept(kept, res.setSize);
        }
        if (par.msaFormatMode == Parameters::FORMAT_MSA_FASTADB_SUMMARY) {
          // gather headers for summary
          for (size_t i = 0; i < res.setSize; i++) {
            if (i == 0) {
              headers.emplace_back(centerSequenceHeader, centerHeaderLength);
            } else if (kept[i] == true) {
              char *header =
                  targetHeaderReader->getData(seqIds[i - 1], thread_idx);
              size_t length = targetHeaderReader->getEntryLen(id) - 1;
              headers.emplace_back(header, length);
            }
          }
          result.append(1, '#');
          result.append(par.summaryPrefix);
          result.append(1, '-');
          result.append(SSTR(queryKey));
          result.append(1, '|');
          result.append(summarizer.summarize(out, headers));
          result.append(1, '\n');
          headers.clear();
        }

        size_t start = 0;
        if (par.skipQuery == true) {
          start = 1;
        }
        for (size_t i = start; i < res.setSize; i++) {
          if (kept[i] == false) {
            continue;
          }

          char *header;
          size_t length;
          if (i == 0) {
            header = centerSequenceHeader;
            length = centerHeaderLength;
          } else {
            size_t id = seqIds[i - 1];
            header = targetHeaderReader->getData(id, thread_idx);
            length = targetHeaderReader->getEntryLen(id) - 1;
          }

          result.append(1, '>');
          result.append(header, length);
          // need to allow insertion in the centerSequence
          for (size_t pos = 0; pos < res.centerLength; pos++) {
            char aa = res.msaSequence[i][pos];
            result.append(
                1,
                ((aa < MultipleAlignment::NAA) ? subMat.num2aa[(int)aa] : '-'));
          }
          result.append(1, '\n');
        }
      } else if (par.msaFormatMode == Parameters::FORMAT_MSA_STOCKHOLM_FLAT) {
        if (isFiltering) {
          filter.filter(res.setSize, res.centerLength,
                        static_cast<int>(par.covMSAThr * 100),
                        static_cast<int>(par.qid * 100), par.qsc,
                        static_cast<int>(par.filterMaxSeqId * 100), par.Ndiff,
                        (const char **)res.msaSequence, false);
          filter.getKept(kept, res.setSize);
        }

        result.append("# STOCKHOLM 1.0\n");
        size_t start = 0;
        if (par.skipQuery == true) {
          start = 1;
          result.append("#=GF ID ");
          result.append(Util::parseFastaHeader(centerSequenceHeader));
          result.append(1, '\n');
        }
        for (size_t i = start; i < res.setSize; i++) {
          if (kept[i] == false) {
            continue;
          }

          char *header;
          if (i == 0) {
            header = centerSequenceHeader;
          } else {
            size_t id = seqIds[i - 1];
            header = targetHeaderReader->getData(id, thread_idx);
          }
          accession = Util::parseFastaHeader(header);

          result.append(accession);
          result.append(1, ' ');
          // need to allow insertion in the centerSequence
          for (size_t pos = 0; pos < res.centerLength; pos++) {
            char aa = res.msaSequence[i][pos];
            result.append(
                1,
                ((aa < MultipleAlignment::NAA) ? subMat.num2aa[(int)aa] : '-'));
          }
          result.append(1, '\n');
        }
        result.append("//\n");
      } else if (isCA3M == true) {
        size_t filteredSetSize = res.setSize;
        if (isFiltering) {
          filteredSetSize = filter.filter(
              res, alnResults, static_cast<int>(par.covMSAThr * 100),
              static_cast<int>(par.qid * 100), par.qsc,
              static_cast<int>(par.filterMaxSeqId * 100), par.Ndiff);
        }
        if (par.formatAlignmentMode == Parameters::FORMAT_MSA_CA3M_CONSENSUS) {
          for (size_t pos = 0; pos < res.centerLength; pos++) {
            if (res.msaSequence[0][pos] == MultipleAlignment::GAP) {
              out->failure("Error in computePSSMFromMSA. First sequence of MSA is not allowed to contain gaps");
            }
          }

          PSSMCalculator::Profile pssmRes = calculator.computePSSMFromMSA(
              filteredSetSize, res.centerLength, (const char **)res.msaSequence,
              par.wg);
          result.append(">consensus_");
          result.append(centerSequenceHeader, centerHeaderLength);
          result.append(pssmRes.consensus);
          result.append("\n;");
        } else {
          result.append(1, '>');
          result.append(centerSequenceHeader, centerHeaderLength);
          // Retrieve the master sequence
          for (int pos = 0; pos < centerSequence.L; pos++) {
            result.append(1, subMat.num2aa[centerSequence.numSequence[pos]]);
          }
          result.append("\n;");
        }

        Matcher::result_t queryAln;
        unsigned int newQueryKey = seqConcat->dbAKeyMap(queryKey);
        queryAln.qStartPos = 0;
        queryAln.dbStartPos = 0;
        queryAln.backtrace =
            std::string(centerSequence.L, 'M');  // only matches
        CompressedA3M::hitToBuffer(refReader->getId(newQueryKey), queryAln,
                                   result);
        for (size_t i = 0; i < alnResults.size(); ++i) {
          unsigned int key = alnResults[i].dbKey;
          unsigned int targetKey = seqConcat->dbBKeyMap(key);
          unsigned int targetId = refReader->getId(targetKey);
          CompressedA3M::hitToBuffer(targetId, alnResults[i], result);
        }
      }
      resultWriter.writeData(result.c_str(), result.length(), queryKey,
                             thread_idx, shouldWriteNullByte);
      result.clear();

      MultipleAlignment::deleteMSA(&res);
      seqSet.clear();
      seqIds.clear();
      alnResults.clear();
    }

    delete[] kept;
  }
  resultWriter.close(true);
  if (shouldWriteNullByte == false) {
    FileUtil::remove(out, resultWriter.getIndexFileName());
  }
  resultReader.close();
  queryHeaderReader.close();
  qDbr.close();
  if (!sameDatabase) {
    if (isCA3M == false) {
      targetHeaderReader->close();
      delete targetHeaderReader;
    }
    tDbr->close();
    delete tDbr;
  }

  if (refReader != NULL) {
    refReader->close();
    delete refReader;
  }
  if (seqConcat != NULL) {
    delete seqConcat;
  }

#ifdef HAVE_MPI
  MPI_Barrier(MPI_COMM_WORLD);
  // master reduces results
  if (BiosnakeMPI::isMaster()) {
    std::vector<std::pair<std::string, std::string>> splitFiles;
    for (int procs = 0; procs < BiosnakeMPI::numProc; procs++) {
      std::pair<std::string, std::string> tmpFile =
          Util::createTmpFileNames(outDb, outIndex, procs);
      splitFiles.push_back(std::make_pair(tmpFile.first, tmpFile.second));
    }
    DBWriter::mergeResults(outDb, outIndex, splitFiles, isCA3M);
  }
#endif
  return EXIT_SUCCESS;
}
